@@ -18,11 +18,7 @@ let push_name x = function
       { x with Genintern.genv = Environ.push_named decl x.Genintern.genv }
   | _ -> x
 
-[%%if coq = "8.19"]
-let push_gdecl (name,_,_,_) x = push_name x name
-[%%else]
 let push_gdecl (name,_,_,_,_) x = push_name x name
-[%%endif]
 
 let push_glob_ctx glob_ctx x =
   List.fold_right push_gdecl glob_ctx x
@@ -47,16 +43,6 @@ let intern_global_constr_ty { Ltac_plugin.Tacintern.genv = env } ~intern_env ?(e
 let intern_global_context { Ltac_plugin.Tacintern.genv = env } ~intern_env ctx =
   Constrintern.intern_context env ~bound_univs:UnivNames.empty_binders intern_env ctx
 
-[%%if coq = "8.19"]
-let intern_one h =
-  let open Constrexpr in
-  match h with
-  | CLocalAssum(nl,Default bk,_) -> nl |> List.map (fun n -> n.CAst.v,bk,None,mkGHole)
-  | CLocalAssum(nl,Generalized(bk,_),_) -> nl |> List.map (fun n -> n.CAst.v,bk,None,mkGHole)
-  | CLocalDef (n,_,None) -> [n.CAst.v,Glob_term.Explicit,None,mkGHole]
-  | CLocalDef (n,_,Some _) -> [n.CAst.v,Glob_term.Explicit,Some mkGHole,mkGHole]
-  | CLocalPattern _ -> nYI "irrefutable pattern in synterp"
-[%%else]
 let intern_one h =
   let open Constrexpr in
   match h with
@@ -65,13 +51,7 @@ let intern_one h =
   | CLocalDef (n,_,_,None) -> [n.CAst.v,None,Glob_term.Explicit,None,mkGHole]
   | CLocalDef (n,_,_,Some _) -> [n.CAst.v,None,Glob_term.Explicit,Some mkGHole,mkGHole]
   | CLocalPattern _ -> nYI "irrefutable pattern in synterp"
-[%%endif]
-
-[%%if coq = "8.19"]
-let drop_relevance x = x
-[%%else]
 let drop_relevance (a,_,c,d,e) = (a,c,d,e)
-[%%endif]
   
 let intern_global_context_synterp (ctx : Constrexpr.local_binder_expr list) : Glob_term.glob_decl list =
   CList.concat_map intern_one ctx |> List.rev
@@ -231,7 +211,7 @@ let univpoly_of ~poly ~cumulative =
   | true, false -> Poly
   | false, _ -> Mono
 
-[%%if coq = "8.19" || coq = "8.20"]
+[%%if coq = "8.20"]
   let of_coq_inductive_definition id =
     let open Vernacentries.Preprocessed_Mind_decl in
     let { flags;  typing_flags; private_ind; uniform; inductives } = id in
@@ -283,7 +263,7 @@ let of_coq_inductive_definition id =
   }
 [%%endif]
 
-[%%if coq = "8.19" || coq = "8.20"]
+[%%if coq = "8.20"]
   let of_coq_record_definition id =
     let open Vernacentries.Preprocessed_Mind_decl in
     let { flags; primitive_proj; kind; records; } : record = id in
@@ -327,20 +307,15 @@ let of_coq_record_definition id =
     name;
     parameters = binders;
     sort;
-    constructor = Some idbuild;
+    constructor = Some idbuild.v;
     fields = cfs;
     univpoly = univpoly_of ~poly ~cumulative
   } 
 [%%endif]
 let intern_record_decl glob_sign (it : raw_record_decl) = glob_sign, it
 
-[%%if coq = "8.19"]
-let mkCLocalAssum x y z = Constrexpr.CLocalAssum(x,y,z)
-let dest_entry (_,_,_,x) = x
-[%%else]
 let mkCLocalAssum x y z = Constrexpr.CLocalAssum(x,None,y,z)
 let dest_entry (_,_,_,_,x) = x
-[%%endif]
 
 let raw_record_decl_to_glob_synterp ({ name; sort; parameters; constructor; fields; univpoly } : raw_record_decl_elpi) : glob_record_decl_elpi =
   let name, space = sep_last_qualid name in
@@ -363,11 +338,7 @@ let raw_record_decl_to_glob_synterp ({ name; sort; parameters; constructor; fiel
         [] fields in
   { name = (space, Names.Id.of_string name); arity; params; constructorname = constructor; fields = List.rev fields; univpoly }
 
-[%%if coq = "8.19"]
-let expr_Type_sort = Glob_term.UAnonymous {rigid=UState.univ_rigid}
-[%%else]
 let expr_Type_sort = Constrexpr_ops.expr_Type_sort
-[%%endif]
 
 let raw_record_decl_to_glob glob_sign ({ name; sort; parameters; constructor; fields; univpoly } : raw_record_decl_elpi) : glob_record_decl_elpi =
   let name, space = sep_last_qualid name in
@@ -453,11 +424,7 @@ let raw_decl_name_to_glob name =
   let name, space = sep_last_qualid name in
   (space, Names.Id.of_string name)
 
-[%%if coq = "8.19"]
-let interp_red_expr = Ltac_plugin.Tacinterp.interp_redexp
-[%%else]
 let interp_red_expr = Redexpr.interp_redexp_no_ltac
-[%%endif]
 
 let raw_constant_decl_to_constr ~depth coq_ctx state { name; typ = (bl,typ); body; red; udecl; atts } =
   let env = coq_ctx.env in
@@ -572,14 +539,15 @@ type raw_ltac_tactic = Ltac_plugin.Tacexpr.raw_tactic_expr
 type glob_ltac_tactic = Ltac_plugin.Tacexpr.glob_tactic_expr
 type top_ltac_tactic = Geninterp.Val.t
 
-type ltac_ty = Int | String | Term | List of ltac_ty
+type ltac_ty = Int | String | Term | OpenTerm | List of ltac_ty
 
 type ('a,'f,'t) t =
   | Int : int            -> ('a,'f,'t) t
   | String : string      -> ('a,'f,'t) t
   | Term : 'a            -> ('a,'f,'t) t
+  | OpenTerm : 'a        -> ('a,'f,'t) t
   | LTac : ltac_ty * 'f  -> ('a,'f,'t) t
-  | LTacTactic : 't  -> ('a,'f,'t) t
+  | LTacTactic : 't      -> ('a,'f,'t) t
 
 type raw =  (raw_term,  raw_ltac_term,  raw_ltac_tactic) t
 type glob = (glob_term, glob_ltac_term, glob_ltac_tactic) t
@@ -597,6 +565,7 @@ let pr_arg f k t x = match x with
   | Int n -> Pp.int n
   | String s -> Pp.qstring s
   | Term s -> f s
+  | OpenTerm s -> f s
   | LTac(_, s) -> k s
   | LTacTactic s -> t s
 
@@ -630,6 +599,7 @@ let glob glob_sign : raw -> _ * glob = function
   | Int _ as x -> glob_sign, x
   | String _ as x -> glob_sign, x
   | Term t -> glob_sign, Term (intern_tactic_constr glob_sign t)
+  | OpenTerm t -> glob_sign, OpenTerm (intern_tactic_constr glob_sign t)
   | LTac(ty,t) -> glob_sign, LTac (ty,fst @@ intern_tactic_constr glob_sign t)
   | LTacTactic t -> glob_sign, LTacTactic (Ltac_plugin.Tacintern.glob_tactic t)
 
@@ -638,6 +608,8 @@ let subst mod_subst = function
   | String _ as x -> x
   | Term t ->
       Term (Ltac_plugin.Tacsubst.subst_glob_constr_and_expr mod_subst t)
+  | OpenTerm t ->
+      OpenTerm (Ltac_plugin.Tacsubst.subst_glob_constr_and_expr mod_subst t)
   | LTac(ty,t) ->
       LTac(ty,(Detyping.subst_glob_constr (Global.env()) mod_subst t))        
   | LTacTactic t ->
@@ -647,6 +619,7 @@ let interp return ist = function
   | Int _ as x -> return x
   | String _ as x -> return x
   | Term t -> return @@ Term(ist,t)
+  | OpenTerm t -> return @@ OpenTerm(ist,t)
   | LTac(ty,v) ->
       let id =
         match DAst.get v with
@@ -654,7 +627,6 @@ let interp return ist = function
         | _ -> assert false in
         return @@ LTac(ty,(ist,id))
   | LTacTactic t -> return @@ LTacTactic (Ltac_plugin.Tacinterp.Value.of_closure ist t)
-  
 
 let add_genarg tag pr_raw pr_glob pr_top glob subst interp =
   let wit = Genarg.make0 tag in
@@ -689,30 +661,30 @@ let mk_indt_decl state univpoly r =
       assert(gls=[]);
       state, E.mkApp uideclc r [up]
 
-let rec do_params_synterp ~depth params k state =
+let rec gparams2lp_synterp ~depth params k state =
   match params with
   | [] -> k state
   | (name,imp,ob,src) :: params ->
       if ob <> None then Coq_elpi_utils.nYI "defined parameters in a record/inductive declaration";
       let src = E.mkDiscard in
-      let state, tgt = do_params_synterp ~depth params k state in
+      let state, tgt = gparams2lp_synterp ~depth params k state in
       let state, imp = in_elpi_imp ~depth state imp in
       state, in_elpi_parameter name ~imp src tgt
       
       
-let rec do_fields_synterp ~depth fields state =
+let rec gfields2lp_synterp ~depth fields state =
   match fields with
   | [] -> state, in_elpi_indtdecl_endrecord ()
   | (f,({ name; is_coercion; is_canonical } as att)) :: fields ->
       let f = E.mkDiscard in
-      let state, fields = do_fields_synterp ~depth fields state in
+      let state, fields = gfields2lp_synterp ~depth fields state in
       in_elpi_indtdecl_field ~depth state att f fields
       
-let do_record_synterp ~depth ~name ~constructorname arity fields state =
+let grecord2lp_synterp ~depth ~name ~constructorname arity fields state =
   let space, record_name = name in
   let qrecord_name = Id.of_string_soft @@ String.concat "." (space @ [Id.to_string record_name]) in
   let arity = E.mkDiscard in
-  let state, fields = do_fields_synterp ~depth fields state in
+  let state, fields = gfields2lp_synterp ~depth fields state in
   let constructor = match constructorname with
     | None -> Name.Name (Id.of_string ("Build_" ^ Id.to_string record_name))
     | Some x -> Name.Name x in
@@ -720,12 +692,12 @@ let do_record_synterp ~depth ~name ~constructorname arity fields state =
 
 let grecord2lp_synterp ~depth state { Cmd.name; arity; params; constructorname; fields; univpoly } =
   let params = List.map drop_relevance params in
-  let state, r = do_params_synterp ~depth params (do_record_synterp ~depth ~name ~constructorname arity fields) state in
+  let state, r = gparams2lp_synterp ~depth params (grecord2lp_synterp ~depth ~name ~constructorname arity fields) state in
   mk_indt_decl state univpoly r
       
-let grecord2lp ~depth state { Cmd.name; arity; params; constructorname; fields; univpoly } =
+let grecord2lp ~loc ~base ~depth state { Cmd.name; arity; params; constructorname; fields; univpoly } =
   let open Coq_elpi_glob_quotation in
-  let state, r = do_params params (do_record ~name ~constructorname arity fields) ~depth state in
+  let state, r = gparams2lp ~loc ~base params ~k:(grecord2lp ~loc ~base ~name ~constructorname arity fields) ~depth state in
   mk_indt_decl state univpoly r
   
 let contract_params env sigma name params nuparams_given t =
@@ -773,39 +745,39 @@ let ginductive2lp_synterp ~depth state { Cmd.finiteness; name; arity; params; nu
   let nuparams = List.map drop_relevance nuparams in
   let params = List.map drop_relevance params in
   let do_constructor ~depth state (name, ty) =
-    let state, ty = do_params_synterp nuparams (fun state -> state, in_elpi_arity E.mkDiscard) ~depth state in
+    let state, ty = gparams2lp_synterp nuparams (fun state -> state, in_elpi_arity E.mkDiscard) ~depth state in
     state, in_elpi_indtdecl_constructor (Name.Name name) ty
   in
   let do_inductive_synterp ~depth state =
     let qindt_name = Id.of_string_soft @@ String.concat "." (space @ [Id.to_string indt_name]) in
-    let state, arity = do_params_synterp nuparams (fun state -> state, in_elpi_arity E.mkDiscard) ~depth state in
+    let state, arity = gparams2lp_synterp nuparams (fun state -> state, in_elpi_arity E.mkDiscard) ~depth state in
     let state, constructors = Coq_elpi_utils.list_map_acc (do_constructor ~depth ) state constructors in
     state, in_elpi_indtdecl_inductive state finiteness (Name.Name qindt_name) arity constructors
   in
-  let state, r = do_params_synterp params (do_inductive_synterp ~depth) ~depth state in
+  let state, r = gparams2lp_synterp params (do_inductive_synterp ~depth) ~depth state in
   mk_indt_decl state univpoly r
   
-let ginductive2lp ~depth state { Cmd.finiteness; name; arity; params; nuparams; nuparams_given; constructors; univpoly } =
+let ginductive2lp ~loc ~depth ~base state { Cmd.finiteness; name; arity; params; nuparams; nuparams_given; constructors; univpoly } =
   let open Coq_elpi_glob_quotation in
   let space, indt_name = name in
   let contract state x =
     let params = List.map drop_relevance params in
     contract_params (get_global_env state) (get_sigma state) indt_name params nuparams_given x in
   let do_constructor ~depth state (name, ty) =
-    let state, ty = do_params nuparams (do_arity (contract state ty)) ~depth state in
+    let state, ty = gparams2lp ~loc ~base nuparams ~k:(garity2lp ~loc ~base (contract state ty)) ~depth state in
     state, in_elpi_indtdecl_constructor (Name.Name name) ty
   in
   let do_inductive ~depth state =
     let short_name = Name.Name indt_name in
     let qindt_name = Id.of_string_soft @@ String.concat "." (space @ [Id.to_string indt_name]) in
-    let state, term_arity = gterm2lp ~depth state (Coq_elpi_utils.mk_gforall arity nuparams) in
-    let state, arity = do_params nuparams (do_arity arity) ~depth state in
-    under_ctx short_name term_arity None (fun ~depth state ->
+    let state, term_arity = gterm2lp ~loc ~depth ~base (Coq_elpi_utils.mk_gforall arity nuparams) state in
+    let state, arity = gparams2lp ~loc ~base nuparams ~k:(garity2lp ~loc ~base arity) ~depth state in
+    under_ctx short_name term_arity None ~k:(fun ~depth state ->
       let state, constructors = Coq_elpi_utils.list_map_acc (do_constructor ~depth ) state constructors in
       state, in_elpi_indtdecl_inductive state finiteness (Name.Name qindt_name) arity constructors, ())
     ~depth state
   in
-  let state, r = do_params params (drop_unit do_inductive) ~depth state in
+  let state, r = gparams2lp ~loc ~base params ~k:(drop_unit do_inductive) ~depth state in
   mk_indt_decl state univpoly r
 
 let in_option = Elpi.(Builtin.option API.BuiltInData.any).API.Conversion.embed
@@ -821,7 +793,7 @@ let ucdeclc = E.Constants.declare_global_symbol "upoly-const-decl"
 
 let gdecl2lp_synterp ~depth state { Cmd.name; params; typ : _; body; udecl } =
   let params = List.map drop_relevance params in
-  let state, typ = do_params_synterp ~depth params (fun state -> state, in_elpi_arity E.mkDiscard) state in
+  let state, typ = gparams2lp_synterp ~depth params (fun state -> state, in_elpi_arity E.mkDiscard) state in
   let body = Option.map (fun _ -> E.mkDiscard) body in
   let name = decl_name2lp name in
   let state, body, gls = in_option ~depth state body in
@@ -832,10 +804,10 @@ let gdecl2lp_synterp ~depth state { Cmd.name; params; typ : _; body; udecl } =
       let state, ud, gls1 = universe_decl.API.Conversion.embed ~depth state ud in
       state, E.mkApp ucdeclc name [body;typ;ud], gls @ gls1
 
-let cdecl2lp ~depth state { Cmd.name; params; typ; body; udecl } =
+let cdecl2lp ~loc ~depth ~base state { Cmd.name; params; typ; body; udecl } =
   let open Coq_elpi_glob_quotation in
-  let state, typ = do_params params (do_arity typ) ~depth state in
-  let state, body = option_map_acc (fun state bo -> gterm2lp ~depth state @@ Coq_elpi_utils.mk_gfun bo params) state body in
+  let state, typ = gparams2lp ~loc ~base params ~k:(garity2lp ~loc ~base typ) ~depth state in
+  let state, body = option_map_acc (fun state bo -> gterm2lp ~loc ~depth ~base (Coq_elpi_utils.mk_gfun bo params) state) state body in
   let name = decl_name2lp name in
   let state, body, gls = in_option ~depth state body in
   match udecl with
@@ -848,18 +820,6 @@ let cdecl2lp ~depth state { Cmd.name; params; typ; body; udecl } =
 let ctxitemc = E.Constants.declare_global_symbol "context-item"
 let ctxendc =  E.Constants.declare_global_symbol "context-end"
 
-[%%if coq = "8.19"]
-let rec do_context_glob_synterp fields ~depth state =
-  match fields with
-  | [] -> state, E.mkGlobal ctxendc
-  | (name,bk,bo,ty) :: fields ->
-      let ty = E.mkDiscard in
-      let bo = Option.map (fun _ -> E.mkDiscard) bo in
-      let state, fields = do_context_glob_synterp fields ~depth state in
-      let state, bo, _ = in_option ~depth state bo in
-      let state, imp = in_elpi_imp ~depth state bk in
-      state, E.mkApp ctxitemc (in_elpi_id name) [imp;ty;bo;E.mkLam fields]
-[%%else]
 let rec do_context_glob_synterp fields ~depth state =
   match fields with
   | [] -> state, E.mkGlobal ctxendc
@@ -870,33 +830,18 @@ let rec do_context_glob_synterp fields ~depth state =
       let state, bo, _ = in_option ~depth state bo in
       let state, imp = in_elpi_imp ~depth state bk in
       state, E.mkApp ctxitemc (in_elpi_id name) [imp;ty;bo;E.mkLam fields]
-[%%endif]
 
-[%%if coq = "8.19"]
-let rec do_context_glob fields ~depth state =
-  match fields with
-  | [] -> state, E.mkGlobal ctxendc
-  | (name,bk,bo,ty) :: fields ->
-      let open Coq_elpi_glob_quotation in
-      let state, ty = gterm2lp ~depth state ty in
-      let state, bo = option_map_acc (gterm2lp ~depth) state bo in
-      let state, fields, () = under_ctx name ty bo (nogls (do_context_glob fields)) ~depth state in
-      let state, bo, _ = in_option ~depth state bo in
-      let state, imp = in_elpi_imp ~depth state bk in
-      state, E.mkApp ctxitemc (in_elpi_id name) [imp;ty;bo;E.mkLam fields]
-[%%else]
-let rec do_context_glob fields ~depth state =
+let rec do_context_glob ~loc fields ~depth ~base state =
   match fields with
   | [] -> state, E.mkGlobal ctxendc
   | (name,_,bk,bo,ty) :: fields ->
       let open Coq_elpi_glob_quotation in
-      let state, ty = gterm2lp ~depth state ty in
-      let state, bo = option_map_acc (gterm2lp ~depth) state bo in
-      let state, fields, () = under_ctx name ty bo (nogls (do_context_glob fields)) ~depth state in
+      let state, ty = gterm2lp ~loc ~depth ~base ty state in
+      let state, bo = option_map_acc (fun state bo -> gterm2lp ~loc ~depth ~base bo state) state bo in
+      let state, fields, () = Coq_elpi_glob_quotation.under_ctx name ty bo ~k:(nogls (do_context_glob ~loc ~base fields)) ~depth state in
       let state, bo, _ = in_option ~depth state bo in
       let state, imp = in_elpi_imp ~depth state bk in
       state, E.mkApp ctxitemc (in_elpi_id name) [imp;ty;bo;E.mkLam fields]
-[%%endif]
 let rec do_context_constr coq_ctx csts fields ~depth state =
   let map s x = constr2lp coq_ctx csts ~depth s (EConstr.of_constr x) in
   match fields with
@@ -908,7 +853,7 @@ let rec do_context_constr coq_ctx csts fields ~depth state =
         | None -> state, None, []
         | Some bo -> let state, bo, gl = map state bo in state, Some bo, gl in
         (* TODO GLS *)
-      let state, fields, gl2 = Coq_elpi_glob_quotation.under_ctx name ty bo (do_context_constr coq_ctx csts fields) ~depth state in
+      let state, fields, gl2 = Coq_elpi_glob_quotation.under_ctx name ty bo ~k:(do_context_constr coq_ctx csts fields) ~depth state in
       let state, bo, gl3 = in_option ~depth state bo in
       let state, imp = in_elpi_imp ~depth state bk in
       state, E.mkApp ctxitemc (in_elpi_id name) [imp;ty;bo;E.mkLam fields], gl0 @ gl1 @ gl2 @ gl3
@@ -916,6 +861,7 @@ let rec do_context_constr coq_ctx csts fields ~depth state =
 
 let strc = E.Constants.declare_global_symbol "str"
 let trmc = E.Constants.declare_global_symbol "trm"
+let open_trmc = E.Constants.declare_global_symbol "open-trm"
 let tacc = E.Constants.declare_global_symbol "tac"
 let intc = E.Constants.declare_global_symbol "int"
 let ctxc = E.Constants.declare_global_symbol "ctx-decl"
@@ -958,14 +904,51 @@ let in_elpi_string_arg ~depth state x =
 let in_elpi_int_arg ~depth state x =
   state, E.mkApp intc (CD.of_int x) [], []
 
-let in_elpi_term_arg ~depth state coq_ctx hyps sigma ist glob_or_expr =
+module NIdSet = struct
+  include Id.Map
+  let counter = ref 0
+  let add x s =
+    if not (mem x s) then begin
+      incr counter;
+      add x !counter s
+    end else
+      s
+
+  let fold f s acc =
+    Id.Map.bindings s |> List.sort (fun (_,n) (_,m) -> m - n) |>
+      List.fold_left (fun acc (x,_) -> f x acc) acc
+
+end
+let free_glob_vars known_vars =
+  let open Glob_term in
+    let rec vars bound vs c = match DAst.get c with
+      | GVar id' -> if Id.Set.mem id' bound then vs else NIdSet.add id' vs
+      | _ -> Glob_ops.fold_glob_constr_with_binders Id.Set.add vars bound vs c in
+    fun rt ->
+      let vs = vars known_vars NIdSet.empty rt in
+      vs
+let close_glob coq_ctx term =
+  let open Glob_term in
+  let fv_set = free_glob_vars coq_ctx.names term in
+  (NIdSet.cardinal fv_set ,NIdSet.fold (fun id t ->
+    DAst.(make (GLambda(Name.Name(id),None,Explicit,mkGHole,t)))) fv_set term)
+
+let in_elpi_term_arg ~loc ~base ~depth state coq_ctx hyps sigma ist glob_or_expr =
   let closure = Ltac_plugin.Tacinterp.interp_glob_closure ist coq_ctx.env sigma glob_or_expr in
   let g = Coq_elpi_utils.detype_closed_glob coq_ctx.env sigma closure in
   let state = Coq_elpi_glob_quotation.set_coq_ctx_hyps state (coq_ctx,hyps) in
-  let state, t = Coq_elpi_glob_quotation.gterm2lp ~depth state g in
+  let state, t = Coq_elpi_glob_quotation.gterm2lp ~loc ~depth ~base g state in
   state, E.mkApp trmc t [], []
- 
-let in_elpi_tac_econstr ~depth ?calldepth coq_ctx hyps sigma state x =
+
+let in_elpi_open_term_arg ~loc ~base ~depth state coq_ctx hyps sigma ist glob_or_expr =
+  let closure = Ltac_plugin.Tacinterp.interp_glob_closure ist coq_ctx.env sigma glob_or_expr in
+  let g = Coq_elpi_utils.detype_closed_glob coq_ctx.env sigma closure in
+  let (n, g) = close_glob coq_ctx g in
+  let state = Coq_elpi_glob_quotation.set_coq_ctx_hyps state (coq_ctx,hyps) in
+  let state, t = Coq_elpi_glob_quotation.gterm2lp ~loc ~depth ~base g state in
+  state, E.mkApp open_trmc (CD.of_int n) [t], []
+
+let in_elpi_tac_econstr ~base:() ~depth ?calldepth coq_ctx hyps sigma state x =
   let state, gls0 = set_current_sigma ~depth state sigma in
   let state, t, gls1 = Coq_elpi_HOAS.constr2lp ~depth ?calldepth coq_ctx E.no_constraints state x in
   state, [E.mkApp trmc t []], gls0 @ gls1
@@ -977,10 +960,10 @@ let in_elpi_elab_term_arg ~depth ?calldepth state coq_ctx hyps sigma ist glob_or
   state, E.mkApp trmc t [], gls0 @ gls1
   
 let singleton (state,x,gls) = state,[x],gls
-let rec in_elpi_ltac_arg ~depth ?calldepth coq_ctx hyps sigma state ty ist v =
+let rec in_elpi_ltac_arg ~loc ~depth ~base ?calldepth coq_ctx hyps sigma state ty ist v =
   let open Ltac_plugin in
   let open Tac in
-  let self ty state = in_elpi_ltac_arg ~depth ?calldepth coq_ctx hyps sigma state ty ist in
+  let self ty state = in_elpi_ltac_arg ~loc ~depth ~base ?calldepth coq_ctx hyps sigma state ty ist in
   let self_list ty state l =
     try
       let state, l, gl = API.Utils.map_acc (self ty) state l in
@@ -999,7 +982,7 @@ let rec in_elpi_ltac_arg ~depth ?calldepth coq_ctx hyps sigma state ty ist v =
   | String ->
       let s = my_cast_to_string v in
       singleton @@ in_elpi_string_arg ~depth state s
-  | Term -> try
+  | Term -> begin try
         let t = Taccoerce.Value.cast (Genarg.topwit Stdarg.wit_open_constr) v in
         let state, t, gls = constr2lp ~depth ?calldepth coq_ctx E.no_constraints state t in
         state, [E.mkApp trmc t []], gls
@@ -1007,7 +990,7 @@ let rec in_elpi_ltac_arg ~depth ?calldepth coq_ctx hyps sigma state ty ist v =
         let closure = Taccoerce.coerce_to_uconstr v in
         let g = Coq_elpi_utils.detype_closed_glob coq_ctx.env sigma closure in
         let state = Coq_elpi_glob_quotation.set_coq_ctx_hyps state (coq_ctx,hyps) in
-        let state, t = Coq_elpi_glob_quotation.gterm2lp ~depth state g in
+        let state, t = Coq_elpi_glob_quotation.gterm2lp ~loc ~depth ~base g state in
         state, [E.mkApp trmc t []], []
       with Taccoerce.CannotCoerceTo _ -> try
         let id = Taccoerce.coerce_to_hyp coq_ctx.env sigma v in
@@ -1015,6 +998,14 @@ let rec in_elpi_ltac_arg ~depth ?calldepth coq_ctx hyps sigma state ty ist v =
         state, [E.mkApp trmc t []], gls
       with Taccoerce.CannotCoerceTo _ ->
         raise (Taccoerce.CannotCoerceTo "a term")
+      end
+  | OpenTerm ->
+      let closure = Taccoerce.coerce_to_uconstr v in
+      let g = Coq_elpi_utils.detype_closed_glob coq_ctx.env sigma closure in
+      let n, g = close_glob coq_ctx g in
+      let state = Coq_elpi_glob_quotation.set_coq_ctx_hyps state (coq_ctx,hyps) in
+      let state, t = Coq_elpi_glob_quotation.gterm2lp ~loc ~depth ~base g state in
+      state, [E.mkApp open_trmc (CD.of_int n) [t]], []
 
 let { CD.cin = of_ltac_tactic; isc = is_ltac_tactic; cout = to_ltac_tactic }, tac = CD.declare {
   CD.name = "ltac1-tactic";
@@ -1029,27 +1020,28 @@ let { CD.cin = of_ltac_tactic; isc = is_ltac_tactic; cout = to_ltac_tactic }, ta
 let in_elpi_ltac_tactic ~depth ?calldepth coq_ctx hyps sigma state t =
   state, [E.mkApp tacc (of_ltac_tactic t) []], []
 
-let in_elpi_tac ~depth ?calldepth coq_ctx hyps sigma state x =
+let in_elpi_tac ~loc ~base ~depth ?calldepth coq_ctx hyps sigma state x =
   let open Tac in
   match x with
   | LTacTactic t -> in_elpi_ltac_tactic ~depth ?calldepth coq_ctx hyps sigma state t
   | LTac(ty,(ist,id)) ->
       let v = try Id.Map.find id ist.Geninterp.lfun with Not_found -> assert false in
       begin try
-        in_elpi_ltac_arg ~depth ?calldepth coq_ctx hyps sigma state ty ist v
+        in_elpi_ltac_arg ~loc ~depth ~base ?calldepth coq_ctx hyps sigma state ty ist v
       with Ltac_plugin.Taccoerce.CannotCoerceTo s ->
         let env = Some (coq_ctx.env,sigma) in
         Ltac_plugin.Taccoerce.error_ltac_variable id env v s end
   | Int x -> singleton @@ in_elpi_int_arg ~depth state x
   | String x -> singleton @@ in_elpi_string_arg ~depth state x
-  | Term (ist,glob_or_expr) -> singleton @@ in_elpi_term_arg ~depth state coq_ctx hyps sigma ist glob_or_expr
+  | Term (ist,glob_or_expr) -> singleton @@ in_elpi_term_arg ~loc ~depth ~base state coq_ctx hyps sigma ist glob_or_expr
+  | OpenTerm (ist,glob_or_expr) -> singleton @@ in_elpi_open_term_arg ~loc ~depth ~base state coq_ctx hyps sigma ist glob_or_expr
 
 let handle_template_polymorphism = function
   | None -> Some false
   | Some false -> Some false
   | Some true -> err Pp.(str "#[universes(template)] is not supported")
 
-[%%if coq = "8.19" || coq = "8.20"]
+[%%if coq = "8.20"]
 let handle_template_polymorphism flags =
   let open Vernacentries.Preprocessed_Mind_decl in
   { flags with template = handle_template_polymorphism flags.template }
@@ -1083,7 +1075,7 @@ let in_elpi_cmd_synterp ~depth ?calldepth state (x : Cmd.raw) =
   | Term raw_term ->
       state, E.mkApp trmc E.mkDiscard [], []
 
-[%%if coq = "8.19" || coq = "8.20"]
+[%%if coq = "8.20"]
 let dest_rdecl raw_rdecl =
   let open Vernacentries.Preprocessed_Mind_decl in
   let { flags = ({ template; poly; cumulative; udecl; finite } as flags); primitive_proj; kind; records } = raw_rdecl in
@@ -1101,7 +1093,7 @@ let interp_structure ~flags udecl kind ~primitive_proj x =
   Record.interp_structure ~flags udecl kind ~primitive_proj x
 [%%endif]
 
-[%%if coq = "8.19" || coq = "8.20"]
+[%%if coq = "8.20"]
 let dest_idecl raw_indt =
   let open Vernacentries.Preprocessed_Mind_decl in
   let { flags = ({ udecl } as flags); typing_flags; uniform; private_ind; inductives } = raw_indt in
@@ -1120,7 +1112,7 @@ let interp_mutual_inductive ~flags ~env ~uniform ~private_ind ?typing_flags ~ude
 [%%endif]
 
 
-let in_elpi_cmd ~depth ?calldepth coq_ctx state ~raw (x : Cmd.top) =
+let in_elpi_cmd ~loc ~depth ~base ?calldepth coq_ctx state ~raw (x : Cmd.top) =
   let open Cmd in
   let hyps = [] in
   match x with
@@ -1128,7 +1120,7 @@ let in_elpi_cmd ~depth ?calldepth coq_ctx state ~raw (x : Cmd.top) =
       let raw_rdecl = of_coq_record_definition raw_rdecl in
       let glob_rdecl = raw_record_decl_to_glob glob_sign raw_rdecl in
       let state = Coq_elpi_glob_quotation.set_coq_ctx_hyps state (coq_ctx,hyps) in
-      let state, t = grecord2lp ~depth state glob_rdecl in
+      let state, t = grecord2lp ~loc ~base ~depth state glob_rdecl in
       state, t, []
   | RecordDecl (_ist,(glob_sign,raw_rdecl)) ->
       let flags, udecl, primitive_proj, kind, records = dest_rdecl raw_rdecl in
@@ -1141,7 +1133,7 @@ let in_elpi_cmd ~depth ?calldepth coq_ctx state ~raw (x : Cmd.top) =
       let raw_indt = of_coq_inductive_definition raw_indt in
       let glob_indt = raw_indt_decl_to_glob glob_sign raw_indt in
       let state = Coq_elpi_glob_quotation.set_coq_ctx_hyps state (coq_ctx,hyps) in
-      let state, t = ginductive2lp ~depth state glob_indt in
+      let state, t = ginductive2lp ~loc ~depth ~base state glob_indt in
       state, t, []
   | IndtDecl (_ist,(glob_sign,raw_indt)) -> 
       let flags, udecl, typing_flags, uniform, private_ind, inductives = dest_idecl raw_indt in
@@ -1156,7 +1148,7 @@ let in_elpi_cmd ~depth ?calldepth coq_ctx state ~raw (x : Cmd.top) =
   | ConstantDecl (_ist,(glob_sign,raw_cdecl)) when raw ->
       let state, glob_cdecl = raw_constant_decl_to_glob glob_sign raw_cdecl state in
       let state = Coq_elpi_glob_quotation.set_coq_ctx_hyps state (coq_ctx,hyps) in
-      cdecl2lp ~depth state glob_cdecl
+      cdecl2lp ~loc ~depth ~base state glob_cdecl
   | ConstantDecl (_ist,(glob_sign,({ name; typ = (bl,_) } as raw_cdecl))) ->
       let state, udecl, typ, body, gls0 =
         raw_constant_decl_to_constr ~depth coq_ctx state raw_cdecl in
@@ -1176,7 +1168,7 @@ let in_elpi_cmd ~depth ?calldepth coq_ctx state ~raw (x : Cmd.top) =
   | Context (_ist,(glob_sign,raw_ctx)) when raw ->
       let glob_ctx = raw_context_decl_to_glob glob_sign raw_ctx in
       let state = Coq_elpi_glob_quotation.set_coq_ctx_hyps state (coq_ctx,hyps) in
-      let state, t = do_context_glob glob_ctx ~depth state in
+      let state, t = do_context_glob ~loc glob_ctx ~depth ~base state in
       state, E.mkApp ctxc t [], []
   | Context (_ist,(glob_sign,raw_ctx)) ->
       let sigma, ctx = ComAssumption.interp_context coq_ctx.env (get_sigma state) raw_ctx in
@@ -1187,7 +1179,7 @@ let in_elpi_cmd ~depth ?calldepth coq_ctx state ~raw (x : Cmd.top) =
   | String x -> in_elpi_string_arg ~depth state x
   | Term (ist,glob_or_expr) when raw ->
       let sigma = get_sigma state in
-      in_elpi_term_arg ~depth state coq_ctx hyps sigma ist glob_or_expr
+      in_elpi_term_arg ~loc ~depth ~base state coq_ctx hyps sigma ist glob_or_expr
   | Term (ist,glob_or_expr) ->
       let sigma = get_sigma state in
       in_elpi_elab_term_arg ~depth ?calldepth state coq_ctx hyps sigma ist glob_or_expr
