@@ -14,11 +14,12 @@ val coq_rest_builtins : BuiltIn.declaration list
 
 val clauses_for_later_interp : (qualified_name * Ast.program * Names.Id.t list * Coq_elpi_utils.clause_scope) list State.component
 
-val set_accumulate_to_db_interp : (((qualified_name * Ast.program * Names.Id.t list * Coq_elpi_utils.clause_scope) list -> unit)) -> unit
-val set_accumulate_text_to_db_interp : ((string list -> string -> Coq_elpi_utils.clause_scope -> unit)) -> unit
+val set_accumulate_to_db_interp : (loc:Loc.t -> (qualified_name * Ast.program * Names.Id.t list * Coq_elpi_utils.clause_scope) list -> unit) -> unit
+val set_accumulate_text_to_db_interp : (loc:Loc.t -> (string list -> string -> Coq_elpi_utils.clause_scope -> unit)) -> unit
 
 (* In tactic mode some APIs are disabled *)
 val tactic_mode : bool State.component
+val base : Compile.program option State.component
 
 val cache_tac_abbrev : code:qualified_name -> name:qualified_name ->unit
 
