@@ -2,6 +2,8 @@
 (* license: GNU Lesser General Public License Version 2.1 or later           *)
 (* ------------------------------------------------------------------------- *)
 
+[%%import "coq_elpi_config.mlh"]
+
 module API = Elpi.API
 module State = API.State
 module Conv = API.Conversion
@@ -62,13 +64,16 @@ let grafting = let open Conv in let open API.AlgebraicData in declare {
     K("after","",A(id,N),
         B (fun x -> (`After,x)),
         M (fun ~ok ~ko -> function (`After,x) -> ok x | _ -> ko ()));
+    K("remove","",A(id,N),
+      B (fun x -> (`Remove,x)),
+      M (fun ~ok ~ko -> function (`Remove,x) -> ok x | _ -> ko ()));
     K("replace","",A(id,N),
       B (fun x -> (`Replace,x)),
       M (fun ~ok ~ko -> function (`Replace,x) -> ok x | _ -> ko ()));
   ]
 } |> CConv.(!<)
 
-type clause = string option * ([ `After | `Before | `Replace ] * string) option * API.Data.term
+type clause = string option * ([ `After | `Before | `Remove | `Replace ] * string) option * API.Data.term
 
 let clause = let open Conv in let open API.AlgebraicData in declare {
   ty = TyName "clause";
@@ -90,7 +95,7 @@ The name and the grafting specification can be left unspecified.|};
 } |> CConv.(!<)
 
 let set_accumulate_to_db_synterp, get_accumulate_to_db_synterp =
-  let f = ref (fun _ -> assert false) in
+  let f = ref (fun ~loc:_ _ -> assert false) in
   (fun x -> f := x),
   (fun () -> !f)
 
@@ -135,13 +140,33 @@ let located = let open Conv in let open API.AlgebraicData in declare {
 let list = B.list
 let pair = B.pair
 let option = B.option
+
+let invocation_site_loc : API.Ast.Loc.t State.component =
+  State.declare_component ~name:"coq-elpi:invocation-site-loc" ~descriptor:interp_state
+  ~pp:(fun fmt x -> Format.fprintf fmt "%a" API.Ast.Loc.pp x)
+  ~init:(fun () -> API.Ast.Loc.initial "(should-not-happen)")
+  ~start:(fun x -> x) ()
+let invocation_site_loc_synterp : API.Ast.Loc.t State.component =
+  State.declare_component ~name:"coq-elpi:invocation-site-loc" ~descriptor:synterp_state
+  ~pp:(fun fmt x -> Format.fprintf fmt "%a" API.Ast.Loc.pp x)
+  ~init:(fun () -> API.Ast.Loc.initial "(should-not-happen)")
+  ~start:(fun x -> x) ()
+
+[%%if elpi >= (1, 20, 0)]
+let compat_graft x = x
+[%%else]
+let compat_graft = Option.map (function `Remove, _ -> nYI "clause removal" | ((`Replace | `Before | `After), _) as x -> x)
+[%%endif]
+
 type accumulation_item = qualified_name * API.Ast.program * Id.t list * Coq_elpi_utils.clause_scope
 let accumulate_clauses ~clauses_for_later ~accumulate_to_db ~preprocess_clause ~scope ~dbname clauses ~depth ~options state =
-  let loc = API.Ast.Loc.initial "(elpi.add_clause)" in
+  let invocation_loc = State.get invocation_site_loc_synterp state in
+  let loc = invocation_loc in
   let dbname = Coq_elpi_utils.string_split_on_char '.' dbname in
   let clauses scope =
    clauses |> CList.rev_map (fun (name,graft,clause) ->
      let vars, clause = preprocess_clause ~depth clause in
+     let graft = compat_graft graft in
      let clause = U.clause_of_term ?name ?graft ~depth loc clause in
      (dbname,clause,vars,scope)) in
   let local = (options : options).local = Some true in
@@ -158,7 +183,7 @@ let accumulate_clauses ~clauses_for_later ~accumulate_to_db ~preprocess_clause ~
   | B.Given CurrentModule ->
        let scope = if local then Local else Regular in
        let f = accumulate_to_db in
-       f (clauses scope);
+       f ~loc:(to_coq_loc loc) (clauses scope);
        state, (), []
 
 let locate_module, locate_module_type =
@@ -240,17 +265,6 @@ let current_path, current_section_path =
      !: (mp2path (Names.ModPath.MPfile sections)))),
   DocAbove)
   
-let invocation_site_loc : API.Ast.Loc.t State.component =
-  State.declare_component ~name:"coq-elpi:invocation-site-loc" ~descriptor:interp_state
-  ~pp:(fun fmt x -> Format.fprintf fmt "%a" API.Ast.Loc.pp x)
-  ~init:(fun () -> API.Ast.Loc.initial "(should-not-happen)")
-  ~start:(fun x -> x) ()
-let invocation_site_loc_synterp : API.Ast.Loc.t State.component =
-  State.declare_component ~name:"coq-elpi:invocation-site-loc" ~descriptor:synterp_state
-  ~pp:(fun fmt x -> Format.fprintf fmt "%a" API.Ast.Loc.pp x)
-  ~init:(fun () -> API.Ast.Loc.initial "(should-not-happen)")
-  ~start:(fun x -> x) ()
-  
 module SynterpAction = struct
   open Declaremods
   type action =
@@ -507,7 +521,11 @@ module SynterpAction = struct
     let open Notation in
     let open CConv in
   
-    [MLData action;
+    [
+      
+     MLData action;
+     MLData Tree.group;
+ 
      MLCode (Pred("coq.synterp-actions",
        Out(list action,"A",
        Read(unit_ctx,"Get the list of actions performed during the parsing phase (aka synterp) up to now.")),
@@ -561,7 +579,7 @@ module SynterpAction = struct
 
   type 'a replay = 'a -> State.t -> State.t * ModPath.t option 
 
-  [%%if coq = "8.19" || coq = "8.20"]
+  [%%if coq = "8.20"]
   let interp_close_section = Lib.Interp.close_section
   [%%else]
   let interp_close_section = Declaremods.Interp.close_section
@@ -932,7 +950,7 @@ let attribute_value = let open API.AlgebraicData in let open CConv in declare {
 
 let attribute = attribute attribute_value
 
-[%%if coq = "8.19" || coq = "8.20"]
+[%%if coq = "8.20"]
 let synterp_close_section = Lib.Synterp.close_section
 [%%else]
 let synterp_close_section = Declaremods.Synterp.close_section
